#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/file.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "settings.h"
#include "misc.h"
#include "crc.h"
#include "cmd.h"


#define LIMIT 8
static char* names[LIMIT];
static int fds[LIMIT];


int
getCmdFd(char* dst)
{
	int i;
	char name[PATH_LEN];

	for(i = 0; i < LIMIT && names[i]; i++)
		if (!strcmp(names[i], dst)) break;

	if (i == LIMIT) {
		message(ERROR, "already opened maximum cmd-files [%d]", LIMIT);
		return -1;
	}

	if (names[i]) return fds[i];

	SNPRINTF(name, PATH_LEN, "%s/%s/todo", CMD, dst);
	if ((fds[i] = open(name, O_CREAT|O_RDWR, 0644)) < 0) {
		message(SYS|ERROR, "cannot open %s", name);
		return -1;
	}
	names[i] = strdup(dst);
	return fds[i];
}

char
putCmd(char* src, char* dst, char* cmd, char* aux)
{
	char command[LONG_STR];
	int fd, l;

	if (!(cmd && dst && src && (fd = getCmdFd(dst)) >= 0))
		return FAILED;

	if (flock(fd, LOCK_EX)) message(SYS|ERROR, "cannot lock %s", dst);

	l = SNPRINTF(command, LONG_STR, "%lu|%s|%s|%s.\n", now, src, cmd, aux?:"");

	lseek(fd, 0, SEEK_END);
	if (write(fd, command, l) < l) message(SYS|ERROR, "cannot write cmd");
	if (flock(fd, LOCK_UN)) message(SYS|ERROR, "cannot unlock %s", dst);
	return OK;
}

char
putDataCmd(char* src, char* dst, char* data, int size, char* cmd)
{
	unsigned crc;
	int l, fd, dataFD;
	char command[LONG_STR];
	char dataName[PATH_LEN];
	char tmp[PATH_LEN];
	int n = -1;

	if (!(cmd && dst && src && data && size > 0 && (fd = getCmdFd(dst)) >= 0))
		return FAILED;

	crc = adler32(1, data, size);

	SNPRINTF(tmp, PATH_LEN, "%s/%s/.tmp.%X", CMD, dst, crc);
	if ((dataFD = open(tmp, O_CREAT|O_WRONLY|O_TRUNC, 0644)) < 0)
		return message(SYS|ERROR, "cannot open data file %s", tmp);
	if (write(dataFD, data, size) < size) {
		close(dataFD);
		unlink(tmp);
		return message(SYS|ERROR, "cannot write %d bytes to file %s", size, tmp);
	}
	close(dataFD);


	if (flock(fd, LOCK_EX)) message(SYS|ERROR, "cannot lock %s", dst);
	do {
		SNPRINTF(dataName, PATH_LEN, "%s/%s/%08X.%d", CMD, dst, crc, ++n);
	} while (exists(dataName));

	if (rename(tmp, dataName)) {
		if (flock(fd, LOCK_UN)) message(SYS|ERROR, "cannot unlock %s", dst);
		return message(SYS|ERROR, "cannot rename %s -> %s", tmp, dataName);
	}

	l = SNPRINTF(command, LONG_STR, "%lu|%s|%s|DATA:%08X.%d:%d.\n",
				 now, src, cmd, crc, n, size);

	lseek(fd, 0, SEEK_END);
	if (write(fd, command, l) < l) message(SYS|ERROR, "cannot write cmd %s", cmd);
	if (flock(fd, LOCK_UN)) message(SYS|ERROR, "cannot unlock %s", dst);
	return OK;
}


Cmd*
getCmd(int fd, char* dst)
{
	if (flock(fd, LOCK_EX)) {
		fprintf(stderr, "cannot lock [%d]", fd);
		return NULL;
	}

	while (TRUE) {
		char str[LONG_STR];
		Cmd* cmd;
		char *eol, *src, *command, *aux;
		int readed;

		if ((readed = read(fd, str, LONG_STR - 1)) < 0) {
			message(SYS|ERROR, "cannot read cmd [%d]", fd);
			break;
		}
		if (readed == 0) break;

		str[readed] = 0;

		if (!(eol = strchr(str, '\n'))) {
			message(ERROR, "too long cmd: %s", str);
			ftruncate(fd, 0);
			break;
		}
		*eol++ = 0;

		if (strlen(str) < 10) {
			message(ERROR, "too short cmd: %s", str);
			ftruncate(fd, 0);
			break;
		}
		if (eol[-2] != '.' && eol[-2] != '+') {
			message(ERROR, "bad magic in cmd: %s", str);
			ftruncate(fd, 0);
			break;
		}

		if (readed == eol - str) {
			ftruncate(fd, 0);
			if (eol[-2] == '+') break;
			else lseek(fd, 0, SEEK_SET);
		} else if (eol[-2] == '+') {
			lseek(fd, (eol - str) - readed, SEEK_CUR);
			continue;
		} else {
			char end = '+';
			lseek(fd, (eol - str) - readed - 2, SEEK_CUR);
			write(fd, &end, 1);
			lseek(fd, 1, SEEK_CUR);
		}
		eol[-2] = 0;
		flock(fd, LOCK_UN);

		if (!(src = strchr(str, '|'))) return NULL;
		*src++ = 0;
		if (!(command = strchr(src, '|'))) return NULL;
		*command++ = 0;
		if (!(aux = strrchr(command, '|'))) return NULL;
		*aux++ = 0;

		if (!(cmd = MALLOC(sizeof(Cmd)))) return NULL;


		cmd->src = strdup(src);
		cmd->cmd = strdup(command);
		if (!strncmp(aux, "DATA:", strlen("DATA:"))) {
			char* fname = aux + strlen("DATA:");
			char* size = strchr(fname, ':');
			if (size) {
				*size++ = 0;
				cmd->size = strtoul(size, NULL, 10);
			}
			if (cmd->size > 0) {
				char dataName[PATH_LEN];
				int fd;
				SNPRINTF(dataName, PATH_LEN, "%s/%s/%s", CMD, dst, fname);
				fd = open(dataName, O_RDONLY);
				if (fd >= 0) {
					cmd->data = MALLOC(cmd->size);
					if (read(fd, cmd->data, cmd->size) < cmd->size) {
						message(SYS|ERROR, "cannot read %s", dataName);
						cmd->data = FREE(cmd->data);
#if CHECK_CRC
					} else if (strtoul(fname, NULL, 16) != adler32(1, cmd->data, cmd->size)) {
						message(ERROR, "crc mismatch for %s", dataName);
						cmd->data = FREE(cmd->data);
#endif
					}
					close(fd);
					if (unlink(dataName))
						message(SYS|ERROR, "cannot unlink %s", dataName);
				} else {
					message(SYS|ERROR, "cannot open %s", dataName);
				}
			}
		} else if (*aux) {
			cmd->aux = strdup(aux);
		}

		return cmd;
	}

	lseek(fd, 0, SEEK_SET);
	flock(fd, LOCK_UN);
	return NULL;
}


Cmd*
freeCmd(Cmd* cmd)
{
	if (cmd) {
		FREE(cmd->data);
		FREE(cmd->aux);
		FREE(cmd->cmd);
		FREE(cmd->src);
		FREE(cmd);
	}
	return NULL;
}

