/*	Copyright (C) 1999 Maxim Zotov

	$Id: strings.c,v 1.16 2002/11/01 22:56:20 maxim Exp $
*/

#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>

#include "misc.h"
#include "settings.h"
#include "strings.h"
#include "charsets.h"
#include "crc.h"
#include "list.h"

#define STORE_STEP 16*1024
#define COUNT_STEP 1000

#define STRINGS_MAGIC 0x23630585
#define DEFAULT_HASH_SIZE 1024


Strings*
newStrings(unsigned hashSize)
{
	Strings* s = MALLOC(sizeof(Strings));
	if (!s) return NULL;
	hashSize = biRound(hashSize, 512);
	if (hashSize > 65536) hashSize = 65536;
	s->hash = MALLOC(hashSize * sizeof(unsigned));
	if (!s->hash) return FREE(s);

	s->hashSize = --hashSize;
	return s;
}

unsigned
string2id(Strings* s, char* str)
{
	unsigned* h = s->hash + (hash(str) & s->hashSize);
	unsigned id = *h;
	unsigned l;
	String* string;
	if (s->map) {
		while (id) {
			String* string;
			if (id >= s->mapSize) {
				message(ERROR, "%s: %d >= %d!!!", str, id, s->mapSize);
				return 0;
			}
			string = (String*)(s->map + id);
			if (!memcmp(str, string->string, string->length)) return id;
			id = string->next;
		}
		return 0;
	}

	while (id) {
		String* string;
		if (id>>12 >= s->nOfBlocks) {
			message(ERROR, "%s: %d >= %d!!!", str, id>>12, s->nOfBlocks);
			return 0;
		}
		string = (String*)((char*)(s->blocks[id>>12]) + (id & 0xFFF));
		if (!memcmp(str, string->string, string->length)) return id;
		id = string->next;
	}
	l = strlen(str);
	if (!s->block || s->block->length + l + 6 > BLOCK_SIZE) {
		StringBlock* block = MALLOC(BLOCK_SIZE);
		if (!block) return 0;

		s->blocks = REALLOC(s->blocks, sizeof(StringBlock*)*(s->nOfBlocks + 1));
		if (!s->blocks) return 0;

		block->length = 2 * sizeof(unsigned) + (s->nOfBlocks ? 0 : 5 * sizeof(unsigned));
		s->blocks[s->nOfBlocks++] = s->block = block;
	}
	s->count++;
	s->changed = YES;
	string = (String*)((char*)(s->block) + s->block->length);

	id = ((s->nOfBlocks - 1) << 12) + s->block->length;

	s->block->length += l + 6;
//	if ((s->block->length & 3)) s->block->length = (s->block->length + 4) & ~3;

	memcpy(string->string, str, ++l);
	string->length = l;
	string->next = *h;
	*h = id;

	return id;
}

char*
id2string(Strings* s, unsigned id)
{
	if (s->map) {
		return id < s->mapSize ? ((String*)(s->map + id))->string : NULL;
	} else {
		String* string = (String*)((char*)(s->blocks[id>>12]) + (id & 0xFFF));
		return string->string;
	}
}

unsigned
string2id_mapped(Strings* s, char* str)
{
	unsigned* h = s->hash + (hash(str) & s->hashSize);
	unsigned id = *h;
	if (!s->map) return 0;
	while (id) {
		String* string;
		if (id >= s->mapSize) {
			message(ERROR, "%s: %d >= %d!!!", str, id, s->mapSize);
			return 0;
		}
		string = (String*)(s->map + id);
		if (!memcmp(str, string->string, string->length)) return id;
		id = string->next;
	}
	return 0;
}

char*
id2string_mapped(Strings* s, unsigned id)
{
	return s->map && id < s->mapSize ? ((String*)(s->map + id))->string : NULL;
}


char
writeStrings(Strings* s, char* file, char force)
{
	unsigned i;
	unsigned buf[5];
	unsigned hashSize;
	int fd;

	if (force) {
		s->readedBlock = 0;
		if (s->count) s->changed = YES;
	}

	if (!s->changed) return OK;

	if ((fd = open(file, O_CREAT|O_RDWR, 0666)) < 0)
		return message(SYS|ERROR, "cannot open %s", file);

	if (flock(fd, LOCK_EX)) {
		message(SYS|ERROR, "cannot lock %s", file);
		close(fd);
		return FAILED;
	}

	if (read(fd, buf, sizeof(buf)) == sizeof(buf)) {
		if (buf[2] != STRINGS_MAGIC || buf[3] != s->hashSize) {
			message(ERROR, "bad file %s (%X|%X, %X|%X)", file,
					buf[2], STRINGS_MAGIC,
					buf[3], s->hashSize);
			s->readedBlock = 0;
		}
	}
	lseek(fd, 0, SEEK_SET);

	hashSize = (s->hashSize + 1) * sizeof(unsigned);
	buf[0] = STRINGS_MAGIC;
	buf[1] = s->hashSize;
	buf[2] = s->nOfBlocks;
	buf[3] = s->count;
	buf[4] = adler32(0, (char*)s->hash, hashSize);

	if (s->readedBlock) {
		StringBlock* b = s->blocks[0];
		debug("writing header");

		memcpy(b->data, buf, sizeof(buf));
		b->crc = adler32(0, b->data, b->length - 2 * sizeof(unsigned));
		write(fd, b, 2 * sizeof(unsigned)  + sizeof(buf));
		lseek(fd, s->readedBlock * BLOCK_SIZE, SEEK_SET);
	} else {
		memcpy(s->blocks[0]->data, buf, sizeof(buf));
	}
	for (i = s->readedBlock; i < s->nOfBlocks; i++) {
		StringBlock* b = s->blocks[i];
		debug("writing %d block (%d bytes)", i, b->length);
		b->crc = adler32(0, b->data, b->length - 2 * sizeof(unsigned));
		if (write(fd, b, BLOCK_SIZE) < BLOCK_SIZE) {
			message(SYS|ERROR, "cannot write %s", file);
			close(fd);
			return FAILED;
		}
	}

	debug("writing hash (%d bytes)", hashSize);
	if (write(fd, s->hash, hashSize) < hashSize) {
		message(SYS|ERROR, "cannot write %s", file);
		close(fd);
		return FAILED;
	}

	if (ftruncate(fd, lseek(fd, 0, SEEK_CUR))) {
		message(SYS|ERROR, "cannot truncate %s", file);
		close(fd);
		return FAILED;
	}

	if (close(fd)) return message(SYS|ERROR, "cannot close %s", file);

	return OK;
}


Strings*
readStrings(char* file, unsigned size)
{
	Strings* s;
	char buf[BLOCK_SIZE];
	unsigned* u = (unsigned*) buf;
#if DO_CRC
	StringBlock* b = (StringBlock*) buf;
#endif
	unsigned hashSize, crc, i;

	int fd = open(file, O_RDONLY, 0666);
	if (fd < 0) return newStrings(size);
//		message(SYS|ERROR, "cannot open %s", file);

	if (flock(fd, LOCK_SH)) {
		message(SYS|ERROR, "cannot lock %s", file);
		close(fd);
		return newStrings(size);
	}

	if (read(fd, buf, BLOCK_SIZE) < BLOCK_SIZE) {
		message(ERROR, "cannot read %s", file);
		close(fd);
		return newStrings(size);
	}

	if (u[2] != STRINGS_MAGIC
#if DO_CRC
		|| b->crc != adler32(0, b->data, b->length - 2 * sizeof(unsigned))
#endif
	   ) {
		message(ERROR, "bad file %s (%X, %X)", file, u[2], u[0]);
		close(fd);
		return newStrings(size);
	}


	s = MALLOC(sizeof(Strings));
	s->hashSize = u[2 + 1];
	s->nOfBlocks = u[2 + 2];
	s->count = u[2 + 3];

	debug("strings %s: %d hash, %d strings, %d blocks", file, s->hashSize + 1, s->count, s->nOfBlocks);


	crc = u[2 + 4];
	hashSize = (s->hashSize + 1) * sizeof(unsigned);

	s->readedBlock = s->nOfBlocks ? s->nOfBlocks - 1 : 0;


	s->hash = MALLOC(hashSize);
	s->blocks = MALLOC(s->nOfBlocks * sizeof(StringBlock*));

	for (i = 0; i < s->nOfBlocks; i++) {
		StringBlock* b = s->blocks[i] = MALLOC(BLOCK_SIZE);
		if (i == 0) {
			memcpy(b, buf, BLOCK_SIZE);
		} else {
			if (read(fd, b, BLOCK_SIZE) < BLOCK_SIZE) {
				message(ERROR, "cannot read %s", file);
				close(fd);
				return newStrings(size);
			}
#if DO_CRC
			if (b->crc != adler32(0, b->data, b->length - 2 * sizeof(unsigned))) {
				message(ERROR, "bad block in file %s (%d: %X)", file, i, b->crc);
				close(fd);
				return newStrings(size);
			}
#endif
		}
		debug("%d block: %d bytes, %X crc", i, b->length, b->crc);
	}
	s->block = s->blocks[s->readedBlock];

	if (read(fd, s->hash, hashSize) < hashSize) {
		message(ERROR, "cannot read %s", file);
		close(fd);
		return newStrings(size);
	}

#if DO_CRC
	if (crc != adler32(0, (char*)s->hash, hashSize)) {
		message(ERROR, "bad hash in file %s (%X)", file, crc);
		close(fd);
		return newStrings(size);
	}
#endif

	if (close(fd)) message(SYS|ERROR, "cannot close %s", file);

	return s;
}

Strings*
mapStrings(char* file)
{
	Strings* s;
	unsigned* u;
	unsigned size;
	void* map;

#if DO_CRC
	StringBlock* b;
	unsigned crc;
#endif

	int fd = open(file, O_RDONLY, 0666);
	if (fd < 0) return newStrings(0);
//		message(SYS|ERROR, "cannot open %s", file);

	if (flock(fd, LOCK_SH)) {
		message(SYS|ERROR, "cannot lock %s", file);
		close(fd);
		return newStrings(0);
	}

	size = fileSize(file);

	if ((map = mmap(NULL, size, PROT_READ, MAP_PRIVATE, fd, 0)) < 0) {
		message(SYS|ERROR, "cannot map %s", file);
		close(fd);
		return newStrings(0);
	}

	u = (unsigned*) map;
#if DO_CRC
	b = (StringBlock*) map;
#endif
	if (u[2] != STRINGS_MAGIC
#if DO_CRC
		|| b->crc != adler32(0, b->data, b->length - 2 * sizeof(unsigned))
#endif
	   ) {
		message(ERROR, "bad file %s (%X, %X)", file, u[2], u[0]);
		munmap(map, size);
		close(fd);
		return newStrings(0);
	}
//	debug("fst block: %d bytes, %X crc", b->length, b->crc);


	s = MALLOC(sizeof(Strings));
	s->hashSize = u[2 + 1];
	s->count = u[2 + 3];

	debug("strings %s: %d hash, %d strings, %d blocks", file, s->hashSize + 1, s->count, s->nOfBlocks);

	s->hash = (unsigned*)(map + size - (s->hashSize + 1) * sizeof(unsigned));

#if DO_CRC
	if (u[2 + 4] != adler32(0, s->hash, (s->hashSize + 1) * sizeof(unsigned))) {
		message(ERROR, "bad hash in file %s (%X)", file, u[2 + 4]);
		munmap(map, size);
		close(fd);
		s->hash = NULL;
		freeStrings(s);
		return newStrings(0);
	}
#endif

	s->map = map;
	s->mapSize = size;
	s->mapFD = fd;

	return s;
}

void
showStrings(Strings* s)
{
	printf("Strings: %d hash, %d strings, %d blocks, %d bytes of last block\n",
		   s->hashSize + 1, s->count, s->nOfBlocks, s->block ? s->block->length : 0);
	if (s->map) {
		unsigned i;
		for (i = 0; i <= s->hashSize; i++) {
			unsigned id = s->hash[i];
			while (id) {
				String* string = (String*)(s->map + id);
				puts(string->string);
				id = string->next;
			}
		}
	} else {
		unsigned i;
		for (i = 0; i <= s->hashSize; i++) {
			unsigned id = s->hash[i];
			while (id) {
				String* string = (String*)((char*)s->blocks[id>>12] + (id & 0xFFF));
				puts(string->string);
				id = string->next;
			}
		}
	}
}


Strings*
freeStrings(Strings* s)
{
	int i;
	if (!s) return NULL;
	if (s->map) {
		munmap(s->map, s->mapSize);
		close(s->mapFD);
	} else {
		for (i = 0; i < s->nOfBlocks; i++) FREE(s->blocks[i]);
		FREE(s->blocks);
		FREE(s->hash);
	}
	return FREE(s);
}


char*
addStringToPool(StringPool* pool, char* string)
{
	int length;
	char* r;
	if (!pool || !string) return NULL;
	length = strlen(string) + 1;
	if (!pool->store || pool->size + length > STORE_STEP) {
		char* newstore = (char*) MALLOC(STORE_STEP);
		if (!newstore) return NULL;
		pool->parts = (char**) REALLOC(pool->parts, sizeof(char*)*(pool->nOfParts + 1));
		if (!pool->parts) return NULL;
		pool->parts[pool->nOfParts++] = pool->store = newstore;
		pool->size = 0;
	}
	r = pool->store + pool->size;
	memcpy(r, string, length);
	pool->size += length;

	if (pool->count % COUNT_STEP == 0)
		pool->strings = (char**) REALLOC(pool->strings, sizeof(char*)*(pool->count + COUNT_STEP));
	pool->strings[pool->count++] = r;

	return r;
}

StringPool*
freeStringPool(StringPool* pool)
{
	int i;
	if (pool) {
		for (i = 0; i < pool->nOfParts; i++)
			FREE(pool->parts[i]);
		FREE(pool->parts);
		FREE(pool->strings);
		FREE(pool);
	}
	return NULL;
}

int
hexDigit(char c)
{
	if ('0' <= c && c <= 'f') {
		if (c <= '9') return c - '0';
		if ('A' <= c) {
			if (c <= 'F') return c - 'A' + 10;
			if ('a' <= c) return c - 'a' + 10;
		}
	}
	return -1;
}

char*
unescape(char* str, char plusIsSpace, char isUTF)
{
	char *c, *d;
	if (!str) return NULL;
	for (c = str, d = str; *c; c++, d++) {
		if (*c == '%') {
			int h, l;
			if (c[1] == 'u') {
				int n = 0;
				int u = 0;
				for (n = 0, c++; n < 4; n++) {
					int h = hexDigit(*++c);
					if (h < 0) {
						c--;
						break;
					}
					u = (u << 4) + h;
				}
				*d = unicode2char(u, WIN) ? : '.';
			} else if ((h = hexDigit(c[1])) >= 0 &&
					   (l = hexDigit(c[2])) >= 0) {
				*d = (h << 4) + l;
				c += 2;
			} else {
				*d = *c;
			}
		} else if (plusIsSpace && *c == '+') {
			*d = ' ';
		} else if (d != c) {
			*d = *c;
		}
	}
	*d = 0;
	if (isUTF) str[recode(str, d - str, UTF, WIN)] = 0;

	return str;
}

char*
unhash(char* str)
{
	char *c, *d;
	if (!str) return NULL;
	for (c = str, d = str; *c; c++, d++) {
		if (c[0] == '&' && c[1] == '#') {
			char* end;
			int u = strtol(c + 2, &end, 10);
			if (*end == ';') {
				*d = unicode2char(u, WIN) ? : '.';
				c = end;
			} else {
				*d = *c;
			}
		} else if (d != c) {
			*d = *c;
		}
	}
	*d = 0;
	return str;
}



char* hexDigits = "0123456789ABCDEF";

char*
escape(char* string, char plusIsSpace)
{
	static char buffer[LONG_STR];
	char* p;
	if (!string || !*string) return string;

	for(p = buffer; *string && p - buffer < LONG_STR - 3; string++) {
		if (plusIsSpace && *string == ' ') {
			*p++ = '+';
		} else if (issafe(*string)) {
			*p++ = *string;
		} else {
			*p++ = '%';
			*p++ = hexDigits[(unsigned char)(*string) >> 4];
			*p++ = hexDigits[(unsigned char)(*string) & 0xF];
		}
	}
	*p++ = 0;

	return buffer;
}

char*
escapeLink(char* string)
{
	static char buffer[LONG_STR];
	char* p;
	if (!string || !*string) return "";

	for(p = buffer; *string && p - buffer < LONG_STR - 3; string++) {
		if (*string != '"' && *string != '<' && *string != '>') {
			*p++ = *string;
		} else {
			*p++ = '%';
			*p++ = hexDigits[(unsigned char)(*string) >> 4];
			*p++ = hexDigits[(unsigned char)(*string) & 0xF];
		}
	}
	*p++ = 0;

	return buffer;
}

char*
quote(char* string)
{
	static char* newstring = NULL;
	char* s = string - 1;
    char* c;
	int cnt = 0;
	while ((s = strpbrk(s + 1, "<>&\""))) cnt++;
	if (cnt == 0) return string;

	c = string;
	s = newstring = REALLOC(newstring, strlen(string) + cnt * 5 + 1);
	while (*c) {
		char* entity;
		int l;
		if (*c == '&')		{entity = "&amp;";	l = strlen("&amp;");}
		else if (*c == '<') {entity = "&lt;";	l = strlen("&lt;");}
		else if (*c == '>') {entity = "&gt;";	l = strlen("&gt;");}
		else if (*c == '"') {entity = "&quot;"; l = strlen("&quot;");}
		else {*s++ = *c++; continue;}

		memcpy(s, entity, l);
		s += l;
		c++;
	}
	*s = 0;

	return newstring;
}

char*
quotes(char* string)
{
	static char* newstring = NULL;
	char* s = string - 1;
	char* c;
	int cnt = 0;
	while ((s = strpbrk(s + 1, "\""))) cnt++;
	if (cnt == 0) return string;

	c = string;
	s = newstring = REALLOC(newstring, strlen(string) + cnt *2 + 1);
	while (*c) {
		char* entity;
		int l;
		if (*c == '"') {entity = "\\\""; l = strlen("\\\"");}
		else {*s++ = *c++; continue;}

		memcpy(s, entity, l);
		s += l;
		c++;
	}
	*s = 0;

	return newstring;
}


void
html_escape(List** list, ...)
{
	va_list ap;
	char** string;

	va_start(ap, list);
	while ((string = va_arg(ap, char**))) {
		char* s = *string - 1;
		int cnt = 0;
		while ((s = strpbrk(s+1, "<>&\""))) cnt++;
		if (cnt) {
			int len = strlen(*string);
			char* c = *string;
			char* s = *string = MALLOC(len + cnt*5 + 1);
			while (*c) {
				char* entity;
				int l;
				if (*c == '&')		{entity = "&amp;";	l = strlen("&amp;");}
				else if (*c == '<') {entity = "&lt;";	l = strlen("&lt;");}
				else if (*c == '>') {entity = "&gt;";	l = strlen("&gt;");}
				else if (*c == '"') {entity = "&quot;"; l = strlen("&quot;");}
				else {*s++ = *c++; continue;}

				memcpy(s, entity, l);
				s += l;
				c++;
			}
			*s = 0;

			*list = addToList(*list, *string, NULL);
		}
	}
	va_end(ap);
}

void
sql_escape(List** list, ...)
{
	va_list ap;
	char** string;

	va_start(ap, list);
	while ((string = va_arg(ap, char**))) {
		char* s = *string - 1;
		int cnt = 0;
		while ((s = strpbrk(s+1, "'\"\\\r\n"))) cnt++;
		if (cnt) {
			int len = strlen(*string);
			char* start = *string;
			char* end;
			char* s = *string = MALLOC(len + cnt + 1);
			while ((end = strpbrk(start, "'\"\\\r\n"))) {
				memcpy(s, start, end - start);
				s += end - start;
				*s++ = '\\';
				*s++ = *end++;
				start = end;
			}
			strcpy(s, start);
			*list = addToList(*list, *string, NULL);
		}
	}
	va_end(ap);
}

void
sql_unescape(char* string, ...)
{
	va_list ap;

	va_start(ap, string);
	do {
		char* s = string;
		char* start = s;
		char* bs;

		while ((bs = strchr(start, '\\'))) {
			if (s != start) memmove(s, start, bs - start);
			s += bs - start;
			*s++ = *++bs;
			start = bs + 1;
		}
		if (s != start) strcpy(s, start);
	} while ((string = va_arg(ap, char*)));
	va_end(ap);
}

void
remove_cr(char* string, ...)
{
	va_list ap;

	va_start(ap, string);
	do {
		char* s = string;
		char* start = s;
		char* cr;

		while ((cr = strchr(start, '\r'))) {
			if (s != start && (cr - start)) memmove(s, start, cr - start);
			s += cr - start;
			start = cr + 1;
		}
		if (s != start) strcpy(s, start);
	} while ((string = va_arg(ap, char*)));
	va_end(ap);
}

char*
cat_strings(char* string, ...)
{
	char *dest, *str;
	int length = 0;
	va_list ap;
	for (str = string, va_start(ap, string); str; str = va_arg(ap, char*))
		length += strlen(str);
	va_end(ap);

	dest = calloc(1, length + 1);

	for (str = string, va_start(ap, string); str; str = va_arg(ap, char*))
		strcat(dest, str);
	va_end(ap);

	return dest;
}

char*
add_strings(char** string, ...)
{
	char *str;
	int length = *string ? strlen(*string) : 0;
	int end = length;
	va_list ap;

	va_start(ap, string);
	while ((str = va_arg(ap, char*)))
		length += strlen(str);
	va_end(ap);

	*string = realloc(*string, length + 1);

	va_start(ap, string);
	while ((str = va_arg(ap, char*))) {
		int l = strlen(str);
		memcpy(*string + end, str, l);
		end += l;
	}
	va_end(ap);

	*(*string + end) = 0;
	return *string;
}

char*
add_string(char** string, char* add)
{
	int length = *string ? strlen(*string) : 0;
	int l;

	if (!add) return *string;

	l = strlen(add);

	*string = realloc(*string, length + l + 1);

	memcpy(*string + length, add, l);
	*(*string + length + l) = 0;
	return *string;
}

int
catstr(char* string, int max_size, ...)
{
	char *str, *s = string;
	int length = 0;
	va_list ap;

	if (!string || max_size <= 0) return 0;

	va_start(ap, max_size);
	while ((str = va_arg(ap, char*))) {
		int l = strlen(str);
		if (!l) continue;
		length += l;
		if (length >= max_size) {
			*string = 0;
			return 0;
		}
		memcpy(s, str, l);
		s += l;
	}
	va_end(ap);

	*s = 0;
	return length;
}

char*
chomp(char* str)
{
	if (str) {
		int e = strlen(str) - 1;
		if (e >= 0 && str[e] == '\n') str[e--] = 0;
		if (e >= 0 && str[e] == '\r') str[e] = 0;
	}
	return str;
}

char*
trimSpaces(char* string)
{
	char* s = string + strlen(string) - 1;
	while (s >= string && isspace(*s)) *s-- = 0;
	s = string;
	while (isspace(*s)) s++;
	return s;
}


char toLower[256];

void
initToLower()
{
	unsigned char upper[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZŨ";
	unsigned char lower[] = "abcdefghijklmnopqrstuvwxyz";
	int i;
	toLower[0] = OK;
	for (i = 1; i < 256; i++) toLower[i] = i;
	for (i = 0; upper[i]; i++) toLower[upper[i]] = lower[i];
}

char
tolower(int c)
{
	if (!*toLower) initToLower();
	return toLower[c];
}

char*
strtolower(unsigned char* s)
{
	unsigned char* o = s;
	if (!*toLower) initToLower();
	for (; *s; s++) *s = toLower[*s];
	return o;
}

char
isalnum(char c)
{
	if ('0' <= c) {
		if (c <= '9') return 1;
		if ('A' <= c) {
			if (c <= 'Z') return 1;
			if ('a' <= c && c <= 'z') return 1;
		}
	}
	return 0;
}

char
issafe(char c)
{
	if (',' <= c) {
		if (c <= ':') return OK;
		if ('A' <= c) {
			if (c <= 'Z') return OK;
			if ('^' <= c && c <= 'z') return OK;
		}
	}
	return FAILED;
}

char
isdigit(char c)
{
	if ('0' <= c && c <= '9') return 1;
	return 0;
}

char
isspace(char c)
{
	return (c == ' ' || c == '\n' || c == '\r' || c == '\t');
}

char
ischar(unsigned char c)
{
	if ('0' <= c) {
		if (c <= '9') return YES;
		if (c >= 0xC0 || c == 0xA3 || c == 0xB3 || c == 0xA8 || c == 0xB8) return YES;
		if ('A' <= c) {
			if (c <= 'Z') return YES;
			if ('a' <= c && c <= 'z') return YES;
		}

	}
	return NO;
}

unsigned
hash(char *s)
{
	unsigned int h = 0;
	/* h = h * 65599 + *s; */
	while(*s) h = (h << 16) + (h << 6) - h + (unsigned char) *s++;
	return h;
}


#define N_OF_STRINGS 20

char*
itoa(int n)
{
	static int call = 0;
	static char* strings;

	unsigned u = abs(n);
	char* str, *c;

	if (!strings && (!(strings = MALLOC(N_OF_STRINGS*32)))) return "";
	call %= N_OF_STRINGS;
	c = str = strings + 32 * call++;

	do {
		*c++ = u%10 + '0';
	} while ((u /= 10));
	if (n < 0) *c++ = '-';
	*c-- = 0;

	{	// reverse string
		char* s = str;
		while (s < c) {
			char t = *s;
			*s++ = *c;
			*c-- = t;
		}
	}

	return str;
}

char*
itos(int n)
{
	static int call = 0;
	static char* strings;

	unsigned u = abs(n);
	int p = 0;
	char* str, *c;

	if (!strings && (!(strings = MALLOC(N_OF_STRINGS*32)))) return "";
	call %= N_OF_STRINGS;
	c = str = strings + 32 * call++;

	do {
		if (++p == 4) {*c++ = ','; p = 1;}
		*c++ = u%10 + '0';
	} while ((u /= 10));
	if (n < 0) *c++ = '-';
	*c-- = 0;

	{	// reverse string
		char* s = str;
		while (s < c) {
			char t = *s;
			*s++ = *c;
			*c-- = t;
		}
	}

	return str;
}

int
string2ip(char* str)
{
	if (str) {
		int ip;
		unsigned char* n = (unsigned char*)&ip;
		int count = 4;
		while (*str && !isdigit(*str)) str++;
		while (1) {
			char* e;
			n[--count] = strtol(str, &e, 10);
			if (e == str) return -1;
			if (count && *e == '.') str = e + 1;
			else break;
		}
		return count ? -1 : ip;
	} else {
		return 0;
	}
}

char*
ip2string(int ip)
{
	char *s;
	unsigned char* n = (unsigned char*)&ip;
	static int call = 0;
	static char* strings = NULL;
	if (!strings && (!(strings = MALLOC(2*16)))) return "";
	call &= 1;
	s = strings + 16 * call++;
	snprintf(s, 16, "%d.%d.%d.%d", n[3], n[2], n[1], n[0]);
	return s;
}

int
reverse(char* s)
{
	int l = strlen(s);
	char* e = s + l - 1;
	while (s < e) {
		char t = *s;
		*s++ = *e;
		*e-- = t;
	}
	return l;
}

char*
copystring(char* string)
{
	char* dest = NULL;
	if (string) {
		int length = strlen(string) + 1;
		dest = (char*) MALLOC(length);
		if (dest) memcpy(dest, string, length);
	}
	return dest;
}


int
split(char* string, char* del, int max_len, ...)
{
	static char* buffer = NULL;
	char* token, *arg;

	va_list ap;
	if (max_len < 1) return FAILED;
	if (!buffer && !(buffer = MALLOC(LONG_STR))) return FAILED;

	va_start(ap, max_len);
	strncpy(buffer, string, LONG_STR - 1);

	token = strtok(buffer, del);
	arg = va_arg(ap, char*);
	while (token && arg) {
		strncpy(arg, token, max_len - 1);
		arg[max_len - 1] = 0;
		token = strtok(NULL, del);
		arg = va_arg(ap, char*);
	}

	while(arg) {
		*arg = 0;
		arg = va_arg(ap, char*);
	}
	va_end(ap);

	return OK;
}



char*
fmt(char* format, ...)
{
	va_list ap;
	static char* string = NULL;

	if (!string && !(string = MALLOC(LONG_STR))) return NULL;

	*string = 0;
	va_start(ap, format);
	vsnprintf(string, LONG_STR, format, ap);
	va_end(ap);
	return string;
}


char* chars = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz-_";
unsigned char* charmap = NULL;


char
initcharmap()
{
	unsigned char i = 0;
	unsigned char* c = chars;
	if ((charmap = MALLOC(256))) while(*c) charmap[*c++] = i++;
	else return FAILED;
	return OK;
}

char*
long2str(unsigned u)
{
	static char str[7];
	int i = 6;
	str[i--] = 0;
	do {
		str[i] = chars[u & 63];
		u >>= 6;
	} while (i--);
	return str;
}

char*
longs2str(unsigned u1, unsigned u2)
{
	static char str[13];
	int i = 12;
	str[i--] = 0;
	do {
		str[i] = chars[u2 & 63];
		u2 >>= 6;
	} while (i-- > 6);

	do {
		str[i] = chars[u1 & 63];
		u1 >>= 6;
	} while (i--);
	return str;
}

long
str2long(char* str)
{
	unsigned u;
	unsigned char* c;
	int i;
	if (!str || (!charmap && !initcharmap())) return 0;
	for (u = i = 0, c = str; i < 6 && *c; i++) u = (u << 6) + charmap[*c++];
	return u;
}



char
checkEmail(char* email)
{
	char* s = email;
	while(isalnum(*s) || *s == '.' || *s == '-' || *s == '_') s++;
	if (s == email || *s++ != '@') return FAILED;
	while(isalnum(*s) || *s == '-') s++;
	if (*s++ != '.' ) return FAILED;
	while(isalnum(*s) || *s == '.' || *s == '-') s++;
	if (*s) return FAILED;
	return OK;
}


char
checkPassword(char* password)
{
	char* s = password;
	while(isalnum(*s)) s++;
	return *s ? FAILED : OK;
}

char
checkNick(char* nick)
{
	char* s = nick;
	while(isalnum(*s) || *s == '_' || *s == '-' || *s == '/') s++;
	return *s ? FAILED : OK;
}

int
stripLast(char* url)
{
	char* s = strrchr(url, '/');
	if (!s) return 0;
	*s = 0;
	return s - url;
}



char*
humanSize(long long size)
{
	static int call = 0;
	static char string[4][SHORT_STR];
	char *s = string[++call % 4];
	long long hsize = size < 0 ? 0 : size;
    int fract = 0;
	char* c = "bkMG";
	while (hsize >= 10000 && c[1]) {
        fract = hsize % 1024;
		hsize = (hsize + 512)/1024;
		c++;
	}
	if (*c == 'b') {
		snprintf(s, SHORT_STR, "%5d", (int)hsize);
        return s;
	}

	if (fract >= 973) fract -= 973;
	if (fract >= 512) hsize--;
	if (hsize < 100) snprintf(s, SHORT_STR, "%2d.%d%c", (int)hsize, (fract*10 + 512)/1024, *c);
	else snprintf(s, SHORT_STR, "%4d%c", (int)hsize, *c);
	return s;
}

char*
humanPeriod(int period)
{
	static int call = 0;
	static char string[4][SHORT_STR];
	char *s = string[++call % 4];
	char c;

	if (period < 120) c = 's';
	else if ((period = (period + 30)/60) < 120) c = 'm';
	else if ((period = (period + 30)/60) < 36) c = 'h';
	else {period = (period + 12)/24; c = 'd';}

	snprintf(s, SHORT_STR, "%d%c", period, c);
	return s;
}

char*
strncpy_(char* dest, char* src, int maxLength)
{
	int l = strlen(src) + 1;
	if (l > maxLength) dest[l = maxLength - 1] = 0;
	return memcpy(dest, src, l);
}
