#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "common/settings.h"
#include "common/strings.h"
#include "common/misc.h"

#include "config.h"
#include "core.h"
#include "files.h"

#define FILES_HASH 1024
#define FILES_STEP 64


typedef struct FileInfo_ FileInfo;

struct FileInfo_ {
	int fd;
	struct stat s;
	char name[PATH_LEN];
	int length;

	int hits;
	int count;
	time_t opened;
	time_t checked;
	time_t freed;

	void* map;

	FileInfo* next;
	FileInfo* prev;
};

static FileInfo** files = NULL;
static FileInfo** filesByFD = NULL;
static FileInfo* freeFileInfo = NULL;
static int count = 0;
static int maxCount = 0;
static int hashSize = 0;
static int hits = 0;
static int opens = 0;


char
initFiles()
{
	if (!files) {
		if (!((hashSize = biRound(config->connections, 64)) &&
			  (files = MALLOC(sizeof(FileInfo*) * hashSize--)) &&
			  (filesByFD = MALLOC(sizeof(FileInfo*) * fdLimit))))
			return FAILED;
		count = maxCount = hits = opens = 0;
	} else {
		if (!(filesByFD = REALLOC(filesByFD, sizeof(FileInfo*) * fdLimit)))
			return FAILED;
	}

	return OK;
}


void
freeFiles()
{
	unsigned h;
	for (h = 0; h <= hashSize; h++) {
		FileInfo* f = files[h];
		while(f) {
			FileInfo* next = f->next;

			debug("freeFiles: close %d:%s (%d hits)", f->fd, f->name, f->hits);
			f->next = freeFileInfo;
			freeFileInfo = f;

			if (close(f->fd))
				message(SYS|ERROR, "cannot close %s", f->name);
			f = next;
		}
	}

	files = FREE(files);
	filesByFD = FREE(filesByFD);
}


void
resetTodayFiles()
{
    unsigned h;
	for (h = 0; h <= hashSize; h++) {
		FileInfo* f = files[h];
		while(f) {
			f->hits = 0;
			f = f->next;
		}
	}
	maxCount = opens = hits = 0;

}


int
getFD(char* name, int length)
{
	FileInfo* f;
	struct stat s;
	unsigned h;
	int fd;

	debug("getFD(%s)", name);
	if (!files) {
		errno = EFAULT;
		return -1;
	}

	h = hash(name) & hashSize;
	f = files[h];

	while(f) {
		if (length == f->length && !memcmp(name, f->name, length)) {
			if (now - f->checked >= config->filesCheck) {
				debug("getFD: need check (last: %s)", time2str(f->checked, TODAY_TIME));
				if (stat(name, &s) < 0) return -1;
				if (s.st_mtime != f->s.st_mtime || s.st_size != f->s.st_size) break;
				f->checked = now;
			}
			hits++;
			f->hits++;
			f->count++;
			return f->fd;
		}
		f = f->next;
	}

	debug("getFD: %sopen it", f ? "re" : "");
	if (!f) if (stat(name, &s) < 0) return -1;


	if ((fd = open(name, O_RDONLY)) < 0) return -1;
	if (fd >= fdLimit) {
		message(ERROR, "cannot open %s: too big fd (%d > %d)", name, fd, fdLimit);
		close(fd);
		errno = EMFILE;
		return -1;
	}


	if (!(f = freeFileInfo)) {
		debug("getFD: allocating new files block");
		freeFileInfo = MALLOC(FILES_STEP * sizeof(FileInfo));
		if (!freeFileInfo) {
			errno = ENOMEM;
			return -1;
		}
		for (f = freeFileInfo; f < freeFileInfo + FILES_STEP - 1; f++)
			f->next = f + 1;
		f = freeFileInfo;
	}
	freeFileInfo = freeFileInfo->next;


	if ((f->next = files[h])) f->next->prev = f;
	files[h] = f;
	f->prev = NULL;

	f->fd = fd;
	f->length = length;
	memcpy(f->name, name, length + 1);
	f->count = 1;
	f->hits = 0;
	f->opened = f->checked = now;
	f->freed = 0;
	memcpy(&(f->s), &s, sizeof(struct stat));

	filesByFD[fd] = f;

	if (++count > maxCount) maxCount = count;
	opens++;

	return fd;
}

struct stat*
getFDStat(int fd)
{
	FileInfo* f = filesByFD[fd];
	return f ? &(f->s) : NULL;
}

int
freeFD(int fd)
{
	FileInfo* f;
	if (fd >= fdLimit) {
		message(ERROR, "strange: freeFD for big fd (%d > %d)", fd, fdLimit);
		return close(fd);
	}

	f = filesByFD[fd];

	if (!f) {
		message(ERROR, "strange: cannot find fd %d in list", fd);
		return close(fd);
	}

	debug("freeFD(%d): %s (%d)", fd, f->name, f->count);
	f->count--;
	f->freed = now;
	return 0;
}


void
cleanupFiles()
{
	unsigned h;
	debug("cleanupFiles");
	for (h = 0; h <= hashSize; h++) {
		FileInfo* f = files[h];
		while(f) {
			FileInfo* next = f->next;
			if (f->count == 0 &&  now - f->freed > config->filesKeep) {
				debug("close %d:%s (%d hits)", f->fd, f->name, f->hits);
				if (f->next) f->next->prev = f->prev;
				if (f->prev) f->prev->next = f->next;
				if (files[h] == f) files[h] = f->next;

				f->next = freeFileInfo;
				freeFileInfo = f;
				--count;

				filesByFD[f->fd] = NULL;
				if (close(f->fd))
					message(SYS|ERROR, "cannot close %s", f->name);
			}
			f = next;
		}
	}
}

int
getFilesStats(char* buffer, int size, char show, char* uri)
{
	int l = SNPRINTF(buffer, size, "\n"
					 "<a href=\"%s%s\">Files pool</a> (current/max/opens/hits): %s / %s / %s / %s\n",
					 uri, show ? "" : "?files",
					 itos(count), itos(maxCount), itos(opens), itos(hits));
	if (show) {
		unsigned h;
		l += SNPRINTF(buffer + l, size - l,
					  "%7s %4s %4s %s\n",
					  "Hits", "Conn", "Time", "Filename");
		for (h = 0; h <= hashSize && size > l; h++) {
			FileInfo* f = files[h];
			while(f && size > l) {
				l += SNPRINTF(buffer + l, size - l,
							  "%7s %4d %4s %s\n",
							  itos(f->hits), f->count,
							  humanPeriod(now - f->opened), f->name);
				f = f->next;
			}
		}
	}
	return l;
}

